#ifndef _LINUX_CRYPTO_H
#define _LINUX_CRYPTO_H

/*
 * include/linux/crypto.h
 *
 * Written by Alexander Kjeldaas <astor@fast.no> 1998-10-13
 *
 * Copyright 1998 by Alexander Kjeldaas. This code is licensed under
 * an X11-like license.  See LICENSE.crypto for details.
 *
 */

#define TRANSFORM_DIGEST 0
#define TRANSFORM_CIPHER 1
#define MAX_TRANSFORM    2

#define DIGEST_NONE        0
#define DIGEST_SUM         1
#define DIGEST_CRC_CCITT16 2
#define DIGEST_CRC_CCITT32 3
#define DIGEST_MD4         4
#define DIGEST_MD5         5
#define DIGEST_SHA1        6
#define MAX_DIGEST        20

#define CIPHER_NONE     0
#define CIPHER_XOR      1
#define CIPHER_DES      2
#define CIPHER_FISH2    3
#define CIPHER_BLOWFISH 4
#define CIPHER_CAST128  5
#define CIPHER_IDEA     6
#define CIPHER_SERPENT  7
#define CIPHER_MARS     8
#define CIPHER_TWOFISH  9
#define CIPHER_SKIPJACK 10
#define CIPHER_RC6      11
#define CIPHER_DES_EDE3 12
#define CIPHER_E2       13
#define CIPHER_CAST256  14
#define CIPHER_DFC      15
#define CIPHER_RIJNDAEL 16
#define CIPHER_RC5      17
#define CIPHER_AES      18
#define MAX_CIPHER	20
#define CIPHER_CBC    0x00010000


#define CIPHER_KEYSIZE_ANY  0xFFFFFFFF
#define CIPHER_KEYSIZE_NONE 0x00000000

#define CIPHER_KEYSIZE_40   0x00000010
#define CIPHER_KEYSIZE_56   0x00000040
#define CIPHER_KEYSIZE_64   0x00000080
#define CIPHER_KEYSIZE_80   0x00000200
#define CIPHER_KEYSIZE_96   0x00000800
#define CIPHER_KEYSIZE_112  0x00002000
#define CIPHER_KEYSIZE_128  0x00008000
#define CIPHER_KEYSIZE_160  0x00080000
#define CIPHER_KEYSIZE_168  0x00100000
#define CIPHER_KEYSIZE_192  0x00800000
#define CIPHER_KEYSIZE_256  0x80000000


#ifdef __KERNEL__

#include <linux/list.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/page.h>
#include <asm/semaphore.h>
#include <asm/spinlock.h>

typedef u32 u4byte;
typedef u8  u1byte;


/* A transform group is a group of transforms that behave in a similar
 * fashion */

struct transform_group {
	int tg_id;
	char *tg_name;  /* "cipher" or "digest" */
	spinlock_t tg_lock;
	struct list_head *tg_head;
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *tg_proc_parent_dir;
	int (*read_proc)(char *page, char **start, off_t off,
			 int count, int *eof, void *data);
#endif
};

/* A transform is something that can be found by id or name. Ciphers
   and digests are types of transforms. */

struct transform_implementation {
	struct list_head t_list;
	int t_id;
	char *t_name;
	struct transform_group *t_group;
#ifdef CONFIG_PROC_FS
	/* keep track of the allocated proc_dir_entry */
	struct proc_dir_entry *t_proc;
#endif
};

/* Cipher data structures */

struct cipher_context;
struct cipher_implementation {
	struct transform_implementation trans;
	int blocksize;         /* in bytes */
	int ivsize;            /* in bytes */
	int key_schedule_size; /* in bytes */
	u32 key_size_mask;     /* bit 0 set = 8 bit, ... , 
				* bit 31 set = 256 bit */

	/*
	 * Encrypt the plaintext pointed to by "in".  Write output to
	 * "out". Size of plaintext is "size".  Output buffer must be
	 * able to hold "size" bytes plus padding necessary to make it
	 * a multiple of the cipher blocksize. size <= 0 is
	 * undefined. Returns 0 on success, non-zero on
	 * failure. Function might sleep.
	 * 
	 */
	int (*encrypt)(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);

	/*
	 * Decrypt the ciphertext pointed to by "in".  Write output to
	 * "out". Size of plaintext is "size".  Input buffer is "size"
	 * bytes plus padding necessary to make it a multiple of the
	 * cipher blocksize. size <= 0 is undefined. Returns 0 on
	 * success, non-zero on failure. Function might sleep.
	 *
	 */
	int (*decrypt)(struct cipher_context *cx,
		       const u8 *in, u8 *out, int size);

	int (*set_key)(struct cipher_context *cx, 
		       unsigned char *key, int key_len);

	/* lock and unlock manage the module use counts */ 
	void (*lock)(void);
	void (*unlock)(void);
};

#define MAX_IV_SIZE  32              /* 32 byte - 256 bit */

struct cipher_context {
	struct cipher_implementation *ci;
	u32 *keyinfo;
	int key_length;              /* in bytes */
	u8 iv[MAX_IV_SIZE];
};


/* Digest data structures */

struct digest_context;
struct digest_implementation {
	struct transform_implementation trans;
	int blocksize;           /* in bytes */
	int working_size;        /* in bytes */

	int (*open)(struct digest_context *cx);
	int (*update)(struct digest_context *cx, u8 *in, int size);
	int (*digest)(struct digest_context *cx, u8 *out);
	int (*close)(struct digest_context *cx, u8 *out);

	/* lock and unlock manage the module use counts */
	void (*lock)(void);
	void (*unlock)(void);
};

struct digest_context {
	struct digest_implementation *di;
	u32 *digest_info;
};


struct transform_implementation *find_transform_by_id(int id, int tgroup);
struct transform_implementation *find_transform_by_name(const char *name, 
							int tgroup);

static inline struct cipher_implementation *find_cipher_by_id(int id) 
{
	return (struct cipher_implementation *)
		find_transform_by_id(id, TRANSFORM_CIPHER);
}

static inline struct cipher_implementation *
find_cipher_by_name(const char *name)
{
	return (struct cipher_implementation *)
		find_transform_by_name(name, TRANSFORM_CIPHER);
}
	
static inline struct digest_implementation *find_digest_by_id(int id)
{
	return (struct digest_implementation *)
		find_transform_by_id(id, TRANSFORM_DIGEST);
}

static inline struct digest_implementation *
find_digest_by_name(const char *name)
{
	return (struct digest_implementation *)
		find_transform_by_name(name, TRANSFORM_DIGEST);
}

int register_transform(struct transform_implementation *ti, int tgroup);
int unregister_transform(struct transform_implementation *ti);

static inline int register_cipher(struct cipher_implementation *ci)
{
	return register_transform((struct transform_implementation *)ci,
				  TRANSFORM_CIPHER);
}

static inline int register_digest(struct digest_implementation *di)
{
	return register_transform((struct transform_implementation *)di,
				  TRANSFORM_DIGEST);
}

static inline int unregister_cipher(struct cipher_implementation *ci)
{
	return unregister_transform((struct transform_implementation *)ci);
}

static inline int unregister_digest(struct digest_implementation *ci)
{
	return unregister_transform((struct transform_implementation *)ci);
}


 
/* System initialization */
extern int cryptoapi_init(void);

/* Cipher implementations */

extern int init_cast256(void);
#define CAST256_KEY_SCHEDULE_SIZE (96*sizeof(u32))
extern  int cast256_set_key(struct cipher_context *cx, unsigned char *key, 
			    int key_len);
extern  int cast256_encrypt(struct cipher_context *cx, 
			    const u8 *in, u8 *out, int size);
extern  int cast256_decrypt(struct cipher_context *cx, 
			    const u8 *in, u8 *out, int size);

extern int init_crypton(void);
#define CRYPTON_KEY_SCHEDULE_SIZE (104*sizeof(u32))
extern  int crypton_set_key(struct cipher_context *cx, unsigned char *key, 
			    int key_len);
extern  int crypton_encrypt(struct cipher_context *cx, 
			    const u8 *in, u8 *out, int size);
extern  int crypton_decrypt(struct cipher_context *cx,
			    const u8 *in, u8 *out, int size);

extern int init_serpent(void);
#define SERPENT_KEY_SCHEDULE_SIZE (140*sizeof(u32))
extern  int serpent_set_key(struct cipher_context *cx, unsigned char *key, 
			    int key_len);
extern  int serpent_encrypt(struct cipher_context *cx,
			    const u8 *in, u8 *out, int size);
extern  int serpent_decrypt(struct cipher_context *cx, 
			    const u8 *in, u8 *out, int size);

extern int init_mars(void);
#define MARS_KEY_SCHEDULE_SIZE (40*sizeof(u32))
extern int mars_set_key(struct cipher_context *cx, unsigned char *key, 
			int key_len);
extern int mars_encrypt(struct cipher_context *cx, 
			const u8 *in, u8 *out, int size);
extern int mars_decrypt(struct cipher_context *cx, 
			const u8 *in, u8 *out, int size);

extern int init_rc5(void);
#define RC5_KEY_SCHEDULE_SIZE (34*sizeof(u32))
extern int rc5_set_key(struct cipher_context *cx, unsigned char *key, 
		       int key_len);
extern int rc5_encrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);
extern int rc5_decrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);

extern int init_rc6(void);
#define RC6_KEY_SCHEDULE_SIZE (44*sizeof(u32))
extern int rc6_set_key(struct cipher_context *cx, unsigned char *key, 
		       int key_len);
extern int rc6_encrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);
extern int rc6_decrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);

extern int init_dfc(void);
#define DFC_KEY_SCHEDULE_SIZE (32*sizeof(u32))
extern int dfc_set_key(struct cipher_context *cx, unsigned char *key, 
		       int key_len);
extern int dfc_encrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);
extern int dfc_decrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);

extern int init_rijndael(void);
#define RIJNDAEL_KEY_SCHEDULE_SIZE ((60+60)*sizeof(u32))
extern int rijndael_set_key(struct cipher_context *cx, unsigned char *key, 
			    int key_len);
extern int rijndael_encrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);
extern int rijndael_decrypt(struct cipher_context *cx, 
		       const u8 *in, u8 *out, int size);

extern int init_blowfish(void);
#define BLOWFISH_KEY_SCHEDULE_SIZE ((18+1024)*sizeof(u32))
extern int blowfish_set_key(struct cipher_context *cx,
                            unsigned char *key, int key_len);
extern int blowfish_encrypt(struct cipher_context *cx,
			    const u8 *in, u8 *out, int size);
extern int blowfish_decrypt(struct cipher_context *cx,
                            const u8 *in, u8 *out, int size);

extern int init_twofish(void);
#define TWOFISH_KEY_SCHEDULE_SIZE ((4*256+8+32)*sizeof(u32))
extern int twofish_set_key(struct cipher_context *cx,
                            unsigned char *key, int key_len);
extern int twofish_encrypt(struct cipher_context *cx,
			    const u8 *in, u8 *out, int size);
extern int twofish_decrypt(struct cipher_context *cx,
                            const u8 *in, u8 *out, int size);

extern int init_idea(void);
#define IDEA_KEY_SCHEDULE_SIZE (104*2)
extern int idea_set_key(struct cipher_context *cx, unsigned char *key, 
			int key_len);
extern int idea_encrypt(struct cipher_context *cx,
			const u8 *in, u8 *out, int size);
extern int idea_decrypt(struct cipher_context *cx,
			const u8 *in, u8 *out, int size);

extern int init_des(void);
#define DES_KEY_SCHEDULE_SIZE (32*sizeof(u32))
extern int des_set_key(struct cipher_context *cx, unsigned char *key, 
		       int key_len);
extern int des_encrypt(struct cipher_context *cx,
		       const u8 *in, u8 *out, int size);
extern int des_decrypt(struct cipher_context *cx,
		       const u8 *in, u8 *out, int size);

extern int init_des_ede3(void);
#define DES_EDE3_KEY_SCHEDULE_SIZE (3*32*sizeof(u32))
extern int des_ede3_set_key(struct cipher_context *cx, unsigned char *key,
			    int key_len);
extern int des_ede3_encrypt(struct cipher_context *cx,
			    const u8 *in, u8 *out, int size);
extern int des_ede3_decrypt(struct cipher_context *cx,
			    const u8 *in, u8 *out, int size);

/* Digest implementations */

extern int init_md5(void);
#define MD5_WORKING_SIZE ((4+2+16)*sizeof(u32))
extern int md5_open(struct digest_context *cx);
extern int md5_update(struct digest_context *cx, u8 *in, int size);
extern int md5_current(struct digest_context *cx, u8 *out);
extern int md5_close(struct digest_context *cx, u8 *out);

extern int init_sha1(void);
#define SHA1_WORKING_SIZE ((5+2+16)*sizeof(u32))
extern int sha1_open(struct digest_context *cx);
extern int sha1_update(struct digest_context *cx, u8 *in, int size);
extern int sha1_current(struct digest_context *cx, u8 *out);
extern int sha1_close(struct digest_context *cx, u8 *out);


/* Utility macros */

#define INIT_CIPHER_BLKOPS(name)		\
        encrypt: name##_encrypt,		\
        decrypt: name##_decrypt

#define INIT_CIPHER_OPS(name)			\
        set_key: name##_set_key,		\
        lock:    name##_lock,			\
        unlock:  name##_unlock

#define INIT_DIGEST_OPS(name)			\
        open:   name##_open,			\
	update: name##_update,			\
	digest: name##_digest,			\
	close:  name##_close,			\
	lock:   name##_lock,			\
	unlock: name##_unlock


#define byte(x, nr) ((unsigned char)((x) >> (nr*8)))

#endif /* __KERNEL__ */

#endif
